#  Copyright © 2025 Bentley Systems, Incorporated
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#      http://www.apache.org/licenses/LICENSE-2.0
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import asyncio
from dataclasses import dataclass
from typing import TYPE_CHECKING, Optional
from uuid import UUID

import evo.logging
from evo.aio import AioTransport
from evo.common import APIConnector, Environment, NoAuth
from evo.common.interfaces import ITransport
from evo.common.utils.cache import Cache
from evo.data_converters.common.exceptions import ConflictingConnectionDetailsError, MissingConnectionDetailsError
from evo.oauth import AuthorizationCodeAuthorizer, ClientCredentialsAuthorizer, OAuthScopes, OIDCConnector
from evo.objects import ObjectAPIClient
from evo.objects.utils.data import ObjectDataClient

if TYPE_CHECKING:
    from evo.notebooks import ServiceManagerWidget

logger = evo.logging.getLogger("data_converters")


@dataclass
class EvoWorkspaceMetadata:
    org_id: str = ""
    workspace_id: str = ""
    client_id: str = ""
    client_secret: str = ""
    user_id: str = ""
    hub_url: str = ""
    redirect_url: str = "http://localhost:3000/signin-oidc"
    oidc_issuer: str = ""
    cache_root: str = "./data/cache"

    def has_authentication_code_params(self) -> bool:
        return bool(self.client_id and self.hub_url and self.oidc_issuer and self.redirect_url)

    def has_client_credentials_params(self) -> bool:
        return bool(self.client_id and self.client_secret and self.oidc_issuer and self.user_id)


@dataclass
class EvoObjectMetadata:
    object_id: UUID
    version_id: Optional[str] = None


async def _authorization_code_authorizer(
    transport: ITransport, metadata: EvoWorkspaceMetadata
) -> AuthorizationCodeAuthorizer:
    authorizer = AuthorizationCodeAuthorizer(
        redirect_url=metadata.redirect_url,
        scopes=OAuthScopes.openid | OAuthScopes.evo_discovery | OAuthScopes.evo_object,
        oidc_connector=OIDCConnector(
            transport=transport,
            oidc_issuer=metadata.oidc_issuer,
            client_id=metadata.client_id,
        ),
    )
    await authorizer.login()

    return authorizer


async def client_credentials_authorizer(
    transport: ITransport, metadata: EvoWorkspaceMetadata
) -> ClientCredentialsAuthorizer:
    authorizer = ClientCredentialsAuthorizer(
        oidc_connector=OIDCConnector(
            transport=transport,
            oidc_issuer=metadata.oidc_issuer,
            client_id=metadata.client_id,
            client_secret=metadata.client_secret,
        ),
        scopes=OAuthScopes.openid
        | OAuthScopes.evo_discovery
        | OAuthScopes.evo_workspace
        | OAuthScopes.evo_object
        | OAuthScopes.evo_file,
    )
    await authorizer.authorize()

    return authorizer


def create_evo_object_service_and_data_client(
    evo_workspace_metadata: Optional[EvoWorkspaceMetadata] = None,
    service_manager_widget: Optional["ServiceManagerWidget"] = None,
) -> tuple[ObjectAPIClient, ObjectDataClient]:
    if evo_workspace_metadata and service_manager_widget:
        raise ConflictingConnectionDetailsError(
            "Please provide only one of EvoWorkspaceMetadata or ServiceManagerWidget."
        )
    elif evo_workspace_metadata:
        return create_service_and_data_client_from_metadata(evo_workspace_metadata)
    elif service_manager_widget:
        return create_service_and_data_client_from_manager(service_manager_widget)
    raise MissingConnectionDetailsError(
        "Missing one of EvoWorkspaceMetadata or ServiceManagerWidget needed to construct an ObjectAPIClient."
    )


def create_service_and_data_client_from_manager(
    service_manager_widget: "ServiceManagerWidget",
) -> tuple[ObjectAPIClient, ObjectDataClient]:
    logger.debug("Creating ObjectAPIClient from ServiceManagerWidget")
    environment = service_manager_widget.get_environment()
    connector = service_manager_widget.get_connector()
    service_client = ObjectAPIClient(environment, connector)
    data_client = service_client.get_data_client(service_manager_widget.cache)

    return service_client, data_client


def create_service_and_data_client_from_metadata(
    metadata: EvoWorkspaceMetadata,
) -> tuple[ObjectAPIClient, ObjectDataClient]:
    logger.debug(
        "Creating evo.objects.ObjectAPIClient and evo.objects.utils.data.ObjectDataClient with "
        f"EvoWorkspaceMetadata={metadata}"
    )

    cache = Cache(root=metadata.cache_root, mkdir=True)
    transport = AioTransport(user_agent="evo-data-converters")
    authorizer = NoAuth

    org_uuid = UUID(metadata.org_id) if metadata.org_id else metadata.org_id
    if metadata.has_client_credentials_params():
        authorizer = asyncio.run(client_credentials_authorizer(transport, metadata))
        hub_connector = APIConnector(
            base_url=metadata.hub_url,
            transport=transport,
            authorizer=authorizer,
            additional_headers={"s2s-org-info": metadata.org_id, "s2s-user-info": metadata.user_id},
        )
    else:
        if metadata.has_authentication_code_params():
            authorizer = asyncio.run(_authorization_code_authorizer(transport, metadata))
        else:
            logger.debug("Skipping authentication due to missing required parameters.")

        hub_connector = APIConnector(base_url=metadata.hub_url, transport=transport, authorizer=authorizer)

    workspace_uuid = UUID(metadata.workspace_id) if metadata.workspace_id else metadata.workspace_id

    environment = Environment(
        hub_url=metadata.hub_url,
        org_id=org_uuid,
        workspace_id=workspace_uuid,
    )
    service_client = ObjectAPIClient(environment, hub_connector)
    data_client = service_client.get_data_client(cache)

    return service_client, data_client



#  Copyright © 2025 Bentley Systems, Incorporated
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#      http://www.apache.org/licenses/LICENSE-2.0
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from typing import Any

import numpy as np
from evo_schemas.components import BoundingBox_V1_0_1, Rotation_V1_1_0
from numpy.typing import NDArray
from scipy.spatial.transform import Rotation


def vertices_bounding_box(vertices: NDArray[Any]) -> BoundingBox_V1_0_1:
    bbox_min = np.min(vertices, axis=0)
    bbox_max = np.max(vertices, axis=0)

    return BoundingBox_V1_0_1(
        min_x=float(bbox_min[0]),
        max_x=float(bbox_max[0]),
        min_y=float(bbox_min[1]),
        max_y=float(bbox_max[1]),
        min_z=float(bbox_min[2]),
        max_z=float(bbox_max[2]),
    )


def grid_bounding_box(
    orig: np.ndarray, transformation_matrix: np.ndarray, grid_extents: np.ndarray
) -> BoundingBox_V1_0_1:
    """
    Calculated bounding box of the grid using origin coordinates, transformation matrix and grid extents.
    - Defines the local corners of the grid based on the grid extents.
    - Transforms these local corners using the transformation matrix and adds the origin coordinates to get the global corners.
    - Computes the minimum and maximum coordinates from the global corners.
    - Returns a BoundingBox_V1_0_1 object with the calculated minimum and maximum coordinates for the x, y, and z dimensions.
    """
    local_corners = np.array(
        [
            [0, 0, 0],
            [grid_extents[0], 0, 0],
            [0, grid_extents[1], 0],
            [0, 0, grid_extents[2]],
            [grid_extents[0], grid_extents[1], 0],
            [grid_extents[0], 0, grid_extents[2]],
            [grid_extents[0], grid_extents[1], grid_extents[2]],
            [grid_extents[0], grid_extents[1], grid_extents[2]],
        ]
    )
    global_corners = np.dot(local_corners, transformation_matrix) + orig
    min_coord = np.min(global_corners, axis=0)
    max_coord = np.max(global_corners, axis=0)
    return BoundingBox_V1_0_1(
        min_x=min_coord[0],
        max_x=max_coord[0],
        min_y=min_coord[1],
        max_y=max_coord[1],
        min_z=min_coord[2],
        max_z=max_coord[2],
    )


class UnsupportedRotation(Exception):
    pass


def check_rotation_matrix(rotation_matrix: np.ndarray, threshold: float = 1e-6) -> None:
    u, v, w = rotation_matrix
    for vector in u, v, w:
        if abs(np.linalg.norm(vector) - 1) > threshold:
            raise UnsupportedRotation("scale")
    for a, b, c in [(u, v, w), (v, w, u), (w, u, v)]:
        if np.dot(np.cross(a, b), c) < 0.0:
            raise UnsupportedRotation("invert")
    for a, b in [(u, v), (v, w), (w, u)]:
        if np.dot(a, b) > threshold:
            raise UnsupportedRotation("skew")


def convert_rotation(rotation: Rotation) -> Rotation_V1_1_0:
    """
    Convert scipy Rotation object into a Geoscience Object rotation.

    The Geoscience Object rotation convention is made up of three rotations:
    - dip azimuth, clockwise rotations around the z-axis, in degrees.
    - dip, clockwise rotations around the x'-axis, in degrees.
    - pitch. clockwise rotations around the z''-axis, in degrees.

    The rotations are intrinsic, i.e. the second and third rotation is applied to the rotated coordinate system.
    The notation x' and z'' is used to indicate that the axes they apply to are rotated from the result of the first and second rotation respectively.
    """
    azimuth, dip, pitch = rotation.as_euler("ZXZ", degrees=True)

    # Rotations returned from scipy are counter-clockwise, so we need to negate them to be clockwise.
    # Also, ensure azimuth and pitch values are in the range [0, 360).
    azimuth = (-azimuth + 360) % 360
    dip = -dip
    pitch = (-pitch + 360) % 360

    # Ensure dip is in the range [0, 180). Do this by flipping the azimuth and pitch by 180 degrees.
    if dip < 0:
        azimuth = (azimuth + 180) % 360
        dip = -dip
        pitch = (pitch + 180) % 360
    return Rotation_V1_1_0(dip_azimuth=azimuth, dip=dip, pitch=pitch)


def get_object_tags(path: str, input_type: str, extra_tags: dict = None) -> dict[str, str]:
    return {
        "Source": f"{path} (via Evo Data Converters)",
        "Stage": "Experimental",
        "InputType": input_type,
        **(extra_tags if extra_tags else {}),
    }



#  Copyright © 2025 Bentley Systems, Incorporated
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#      http://www.apache.org/licenses/LICENSE-2.0
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import asyncio
import json
import tempfile
import time
from http import HTTPStatus
from pathlib import Path
from typing import Optional
from urllib.parse import urlencode

import pyarrow as pa
import pyarrow.parquet as pq
import requests

import evo.logging
from evo.common import APIConnector, Environment, HTTPHeaderDict
from evo.data_converters.common.exceptions import ResponseError

logger = evo.logging.getLogger("data_converters")


class BlockSyncClient:
    def __init__(self, environment: Environment, api_connector: APIConnector):
        self.environment = environment
        self.api_connector = api_connector
        self.hub_url = self.environment.hub_url
        self.org_id = self.environment.org_id
        self.workspace_id = self.environment.workspace_id

    def get_auth_header(self) -> HTTPHeaderDict:
        """Get the authorisation header for all Block Model API requests.

        :return: The authorisation headers in dictionary form.
        """
        header_dict: HTTPHeaderDict = asyncio.run(self.api_connector._authorizer.get_default_headers())
        header_dict["API-Preview"] = "opt-in"  # This must be set in the headers to support CRS
        return header_dict

    def check_job_status(self, job_url: str, max_retries: int = 200, retry_delay: int = 1) -> requests.Response:
        """Check the status of a job by making HTTP GET requests to the job URL.

        :param job_url: The URL of the job.
        :param max_retries: (optional) The maximum number of retries. Defaults to 200.
        :param retry_delay: (optional) The delay between retries in seconds. Defaults to 1.

        :return: The response object of the last request if the job succeeded.

        Raises an exception if the job failed.
        """
        retry_count = 0
        has_succeeded = False
        auth_header = self.get_auth_header()

        while retry_count < max_retries:
            response = requests.get(job_url, headers=auth_header)
            response_output = json.dumps(response.json(), indent=4, sort_keys=True)
            logger.info(response_output)

            if response.status_code == HTTPStatus.OK:
                if response.json()["job_status"] == "COMPLETE":
                    logger.info("Job completed successfully!")
                    has_succeeded = True
                    break
                elif response.json()["job_status"] == "FAILED":
                    logger.warning("Job failed!")
                    break
                else:
                    logger.info("Job still in progress...")
                    time.sleep(retry_delay)
                    retry_count += 1

        if not has_succeeded:
            raise ResponseError(f"Job failed: \n Status: {response.status_code} \n Response: {response_output}")
        else:
            return response

    def create_request(self, body: dict) -> str:
        """Sends a create API request to the Block Model API.
        It creates an empty block model which is defined by the body param.

        :param body: The body of the API request which contains information about the model.

        :return: The newly created block model's ID.

        If problems are encountered while accessing the API, these will be raised as exceptions.
        """
        # Compose the URL
        url = f"{self.hub_url}/blockmodel/orgs/{self.org_id}/workspaces/{self.workspace_id}/block-models"

        # Make a POST request and include the JSON body and headers
        auth_header = self.get_auth_header()
        response = requests.post(url=url, json=body, headers=auth_header)

        if response.status_code != HTTPStatus.CREATED:
            raise ResponseError(f"Request failed: \n Status: {response.status_code} \n Response: {response.content!r}")

        response_output = json.dumps(response.json(), indent=4, sort_keys=True)
        response_dict = json.loads(response_output)
        self.check_job_status(job_url=response_dict["job_url"], max_retries=200, retry_delay=1)
        logger.info(f"Success, {body['name']} was created.")
        logger.info("Job url: ", response_dict["job_url"])

        bm_uuid: str = response_dict["bm_uuid"]
        return bm_uuid

    def add_columns_request(self, block_model_id: str, body: dict) -> tuple[str, str]:
        """Sends an update columns API request to the Block Model API.

        It updates a block model with the new columns as per the body param.

        :param block_model_id: The id of the block model to be added to.
        :param body: The body of the API request which contains information about the model.

        :return: The URL of the block model job to add columns, and the URL where the column
        data will be uploaded to.

        If problems are encountered while accessing the API, these will be raised as exceptions.
        """
        # Compose the URL and JSON body
        url = f"{self.hub_url}/blockmodel/orgs/{self.org_id}/workspaces/{self.workspace_id}/block-models/{block_model_id}/blocks"

        # This PATCH request prepares the BlockSync API for data upload
        auth_header = self.get_auth_header()
        response = requests.patch(url=url, json=body, headers=auth_header)

        if response.status_code != HTTPStatus.ACCEPTED:
            raise ResponseError(f"Request failed: \n Status: {response.status_code} \n Response: {response.content!r}")

        response_output = json.dumps(response.json(), indent=4, sort_keys=True)
        response_dict: dict[str, str] = json.loads(response_output)

        job_url: str = response_dict["job_url"]
        upload_url: str = response_dict["upload_url"]

        return job_url, upload_url

    def get_blockmodel_request(self, block_model_id: str) -> requests.Response:
        """Get the BlockModel with ID block_model_id in the current workspace.

        On success, the API will respond with HTTP status code 200 OK along with the BlockModel object in the response body.

        If the BlockModel has been soft-deleted and the query parameter deleted=true has not been provided, the API will respond with status code 410 Gone.

        :param block_model_id: The id of the block model to be retrieved.

        :raises Exception: If there is a problem with the request.
        """
        # Compose the URL and JSON body
        url = (
            f"{self.hub_url}/blockmodel/orgs/{self.org_id}/workspaces/{self.workspace_id}/block-models/{block_model_id}"
        )

        # Make a GET request and include the auth headers
        auth_header = self.get_auth_header()
        logger.info(f"Requesting blockmodel from {url}")
        response = requests.get(url=url, headers=auth_header)

        if response.status_code != HTTPStatus.OK:
            raise ResponseError(f"Request failed: \n Status: {response.status_code} \n Response: {response.content!r}")

        return response

    def complete_blockmodel_upload(self, job_url: str) -> None:
        """Send a POST request to notify the BlockSync API that no more data will be uploaded.

        This will move the job from queued to in progress.

        :param job_url: The job URL to move up the queue.
        """
        auth_header = self.get_auth_header()
        url = f"{job_url}/uploaded"

        # Make a POST request, including the headers
        response = requests.post(url=url, headers=auth_header)
        if response.status_code != HTTPStatus.CREATED:
            raise ResponseError(f"Request failed: \n Status: {response.status_code} \n Response: {response.content!r}")

        response_output = json.dumps(response.json(), indent=4, sort_keys=True)
        response_dict = json.loads(response_output)

        # Check if job completed
        self.check_job_status(job_url=response_dict["job_url"], max_retries=200, retry_delay=1)
        logger.info(f"Job {job_url} complete")

    def upload_parquet(self, upload_url: str, table: pa.Table) -> None:
        """Sends a put request containing a parquet file to the Block Model API.

        :param upload_url: The URL to upload the parquet file data to.
        :param table: The block model data in table form.

        If problems are encountered while accessing the API, these will be raised as exceptions.
        """
        # Compose the headers
        # NOTE: The Seequent ID access token is not required for this request
        headers = {"Content-Type": "application/binary", "x-ms-blob-type": "BlockBlob"}

        # Create a temporary parquet file to upload to BlockSync
        with tempfile.TemporaryDirectory() as tempdirname:
            temp_dir = Path(tempdirname)
            file_name = temp_dir / "upload_file.parquet"
            pq.write_table(table=table, where=file_name)
            # Make a PUT request - include the binary data and headers
            with open(file_name, "rb") as data_stream:
                response = requests.put(url=upload_url, data=data_stream, headers=headers)

        if response.status_code != HTTPStatus.CREATED:
            raise ResponseError(f"Request failed: \n Status: {response.status_code} \n Response: {response.content!r}")

    def get_blockmodel_versions(
        self, block_model_id: str, offset: int = 0, filter_param: Optional[str] = None
    ) -> requests.Response:
        auth_header = self.get_auth_header()
        url = f"{self.hub_url}/blockmodel/orgs/{self.org_id}/workspaces/{self.workspace_id}/block-models/{block_model_id}/versions"

        params = {"offset": str(offset)}
        if filter_param:
            params["filter"] = str(filter_param)

        logger.info(f"Requesting blockmodel versions from {url}")
        response = requests.get(url=f"{url}?{urlencode(params)}", headers=auth_header)

        if response.status_code != HTTPStatus.OK:
            raise ResponseError(f"Request failed: \n Status: {response.status_code} \n Response: {response.content!r}")

        return response

    def get_blockmodel_columns_job_url(self, block_model_id: str, version_uuid: Optional[str] = None) -> str:
        # Compose the URL, query parameters and JSON body
        url = f"{self.hub_url}/blockmodel/orgs/{self.org_id}/workspaces/{self.workspace_id}/block-models/{block_model_id}/blocks"

        body = {"columns": ["*"], "geometry_columns": "indices"}
        if version_uuid:
            body["version_uuid"] = version_uuid

        # Make a POST request and include the JSON body and headers
        auth_header = self.get_auth_header()
        logger.info(f"Requesting blockmodel columns from {url}")
        response = requests.post(url=url, json=body, headers=auth_header)

        # Copy the `job_url` parameter from the response
        if response.status_code != HTTPStatus.OK:
            raise ResponseError(f"Request failed: \n Status: {response.status_code} \n Response: {response.content!r}")

        return str(response.json()["job_url"])

    def get_blockmodel_columns_download_url(self, job_url: str) -> str:
        logger.info(f"Requesting blockmodel columns download URL from {job_url}")
        response = self.check_job_status(job_url=job_url)

        if response.status_code != HTTPStatus.OK:
            raise ResponseError(f"Request failed: \n Status: {response.status_code} \n Response: {response.content!r}")

        return str(response.json()["payload"]["download_url"])

    def download_parquet(self, download_url: str) -> str:
        logger.info(f"Requesting blockmodel parquet file from {download_url}")
        response = requests.get(url=download_url)

        if response.status_code != 200:
            raise ResponseError(f"Request failed: \n Status: {response.status_code} \n Response: {response.content!r}")

        temp_file = tempfile.NamedTemporaryFile(suffix=".parquet", delete=False)
        temp_file.write(response.content)
        temp_file.close()

        return temp_file.name



#  Copyright © 2025 Bentley Systems, Incorporated
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#      http://www.apache.org/licenses/LICENSE-2.0
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

"""
Utility functions for manipulating Evo Hexahedrons
"""

import numpy as np
import numpy.typing as npt
import pyarrow as pa
from evo_schemas.components import Hexahedrons_V1_1_0_Indices as HexahedronsIndices
from evo_schemas.components import Hexahedrons_V1_1_0_Vertices as HexahedronsVertices
from evo_schemas.components import OneOfAttribute_V1_1_0 as OneOfAttribute

from evo.objects.utils.data import ObjectDataClient


def build_vertices(vertices: npt.NDArray[np.float64], data_client: ObjectDataClient) -> HexahedronsVertices:
    """Build a HexahedronsVertices object containing the grid vertices.

    :param vertices: an n x 3 array of cell vertex coordinates (x, y, z)
    :param data_client: ObjectDataClient used to create the HexahedronsVertices object

    :return: An Evo HexahedronsVertices object

    """
    schema = pa.schema(
        [
            ("x", pa.float64()),
            ("y", pa.float64()),
            ("z", pa.float64()),
        ]
    )
    table = pa.Table.from_arrays([vertices[:, 0], vertices[:, 1], vertices[:, 2]], schema=schema)
    go = data_client.save_table(table)
    return HexahedronsVertices.from_dict(go)


def build_indices(
    indices: npt.NDArray[np.intp], data_client: ObjectDataClient, attributes: OneOfAttribute
) -> HexahedronsIndices:
    """Build an Evo HexahedronIndices, containing the indexes for the cell vertices
    and the associated attributes

    :param indices: a n by 8 array of indexes into vertices ,specifying the 8 corner points of each cell
    :param data_client: ObjectDataClient used to create the HexahedronsIndices object
    :param attributes: List of attributes

    :return: An Evo HexahedronsIndices object

    """
    schema = pa.schema(
        [
            ("n0", pa.uint64()),
            ("n1", pa.uint64()),
            ("n2", pa.uint64()),
            ("n3", pa.uint64()),
            ("n4", pa.uint64()),
            ("n5", pa.uint64()),
            ("n6", pa.uint64()),
            ("n7", pa.uint64()),
        ]
    )
    table = pa.Table.from_arrays(
        [
            indices[:, 0],
            indices[:, 1],
            indices[:, 2],
            indices[:, 3],
            indices[:, 4],
            indices[:, 5],
            indices[:, 6],
            indices[:, 7],
        ],
        schema=schema,
    )
    go = data_client.save_table(table)
    hi = HexahedronsIndices.from_dict(go)
    hi.attributes = attributes
    return hi



#  Copyright © 2025 Bentley Systems, Incorporated
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#      http://www.apache.org/licenses/LICENSE-2.0
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import asyncio

import nest_asyncio
from evo_schemas.components import BaseSpatialDataProperties_V1_0_1

import evo.logging
from evo.objects import ObjectAPIClient
from evo.objects.data import ObjectMetadata
from evo.objects.utils import ObjectDataClient

from .generate_paths import generate_paths

logger = evo.logging.getLogger("data_converters")


def publish_geoscience_objects(
    object_models: list[BaseSpatialDataProperties_V1_0_1],
    object_service_client: ObjectAPIClient,
    data_client: ObjectDataClient,
    path_prefix: str = "",
) -> list[ObjectMetadata]:
    """
    Publishes a list of Geoscience Objects.
    """
    objects_metadata = []
    paths = generate_paths(object_models, path_prefix)

    nest_asyncio.apply()

    logger.debug(f"Preparing to publish {len(object_models)} objects to paths: {paths}")
    for obj, obj_path in zip(object_models, paths):
        object_metadata = asyncio.run(publish_geoscience_object(obj_path, obj, object_service_client, data_client))
        logger.debug(f"Got object metadata: {object_metadata}")
        objects_metadata.append(object_metadata)

    return objects_metadata


async def publish_geoscience_object(
    path: str,
    object_model: BaseSpatialDataProperties_V1_0_1,
    object_service_client: ObjectAPIClient,
    data_client: ObjectDataClient,
) -> ObjectMetadata:
    """
    Publish a single Geoscience Object
    """
    logger.debug(f"Publishing Geoscience Object: {object_model}")
    await data_client.upload_referenced_data(object_model.as_dict())
    object_metadata = await object_service_client.create_geoscience_object(path, object_model.as_dict())
    return object_metadata



#  Copyright © 2025 Bentley Systems, Incorporated
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#      http://www.apache.org/licenses/LICENSE-2.0
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from collections import defaultdict
from pathlib import PurePosixPath

from evo_schemas.components import BaseSpatialDataProperties_V1_0_1


def generate_paths(object_models: list[BaseSpatialDataProperties_V1_0_1], path_prefix: str = "") -> list[str]:
    """
    Generates a list of paths where each object will be uploaded to.

    The path for each object follows the pattern of: "<path_prefix>/<object_name>{_<n>}.json"

    For example: "myproject/mysite/myobject_2.json"
    """
    count: defaultdict[str, int] = defaultdict(int)
    paths: list[str] = []

    for obj in object_models:
        obj_path = obj.name + ".json"

        if (n := count[obj.name]) > 0:
            if n == 1:
                # must rename the existing path
                paths[paths.index(obj_path)] = obj.name + "_1.json"

            obj_path = f"{obj.name}_{n + 1}.json"

        paths.append(obj_path)
        count[obj.name] += 1

    if path_prefix:
        # prepend in-place
        for i, obj_path in enumerate(paths):
            paths[i] = str(PurePosixPath(path_prefix, obj_path)).lstrip("/")

    return paths
